/*
 * Device driver for the SYMBIOS/LSILOGIC 53C8XX and 53C1010 family 
 * of PCI-SCSI IO processors.
 *
 * Copyright (C) 1999-2001  Gerard Roudier <groudier@club-internet.fr>
 *
 * This driver is derived from the Linux sym53c8xx driver.
 * Copyright (C) 1998-2000  Gerard Roudier
 *
 * The sym53c8xx driver is derived from the ncr53c8xx driver that had been 
 * a port of the FreeBSD ncr driver to Linux-1.2.13.
 *
 * The original ncr driver has been written for 386bsd and FreeBSD by
 *         Wolfgang Stanglmeier        <wolf@cologne.de>
 *         Stefan Esser                <se@mi.Uni-Koeln.de>
 * Copyright (C) 1994  Wolfgang Stanglmeier
 *
 * Other major contributions:
 *
 * NVRAM detection and reading.
 * Copyright (C) 1997 Richard Waltham <dormouse@farsrobt.demon.co.uk>
 *
 *-----------------------------------------------------------------------------
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef SYM_GLUE_H
#define SYM_GLUE_H

/*
 *  Configuration addendum for FreeBSD.
 *  Low priority probe map.
 * 
 *  This option is used as a bitmap to tell the driver 
 *  about chips that are to be claimed with a low priority 
 *  (-2000) by the probe method. This allows any other driver 
 *  that may return some higher priority value for the same 
 *  chips to take precedence over this driver (sym).
 *  This option may be used when both the ncr driver and this 
 *  driver are configured.
 *
 *  Bits are to be coded as follows:
 *    0x01  ->  810a, 860
 *    0x02  ->  825a, 875, 885, 895
 *    0x04  ->  895a, 896, 1510d
 *    0x08  ->  1010
 *    0x40  ->  810, 815, 825
 *
 *  For example, value 5 tells the driver to claim support 
 *  for 810a, 860, 895a, 896 and 1510d with low priority, 
 *  allowing the ncr driver to take precedence if configured.
 */
#ifndef SYM_SETUP_LP_PROBE_MAP
#define SYM_SETUP_LP_PROBE_MAP 0
#endif

#include <sys/param.h>

/*
 *  Only use the BUS stuff for PCI under FreeBSD 4 and later versions.
 *  Note that the old BUS stuff also works for FreeBSD 4 and spares 
 *  about 1 KB for the driver object file.
 */
#if 	__FreeBSD_version >= 400000
#define	FreeBSD_Bus_Dma_Abstraction
#define	FreeBSD_Bus_Io_Abstraction
#define	FreeBSD_Bus_Space_Abstraction
#endif

/* #define CAM_NEW_TRAN_CODE */

#ifdef  CAM_NEW_TRAN_CODE
#define	FreeBSD_New_Tran_Settings
#endif

/*
 *  Driver configuration options.
 */
#include "opt_sym.h"
#include <dev/sym/sym_conf.h>

#ifndef FreeBSD_Bus_Io_Abstraction
#include "ncr.h"	/* To know if the ncr has been configured */
#endif

#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#ifdef FreeBSD_Bus_Io_Abstraction
#include <sys/module.h>
#include <sys/bus.h>
#endif

#include <sys/proc.h>

#include <pci/pcireg.h>
#include <pci/pcivar.h>

#ifdef	FreeBSD_Bus_Space_Abstraction
#include <machine/bus_memio.h>
/*
 *  Only include bus_pio if needed.
 *  This avoids bus space primitives to be uselessly bloated 
 *  by out-of-age PIO operations.
 */
#ifdef	SYM_CONF_IOMAPPED
#include <machine/bus_pio.h>
#endif
#endif
#include <machine/bus.h>

#ifdef FreeBSD_Bus_Io_Abstraction
#include <machine/resource.h>
#include <sys/rman.h>
#endif

/* For DELAY() that moved to systm.h */
#if __FreeBSD_version < 500000
#include <machine/clock.h>
#endif

/* #ifdef	FreeBSD_New_Tran_Settings */
#if 0
#include <NewCam/cam.h>
#include <NewCam/cam_ccb.h>
#include <NewCam/cam_sim.h>
#include <NewCam/cam_xpt_sim.h>
#include <NewCam/cam_debug.h>
#include <NewCam/scsi/scsi_all.h>
#include <NewCam/scsi/scsi_message.h>
#else
#include <cam/cam.h>
#include <cam/cam_ccb.h>
#include <cam/cam_sim.h>
#include <cam/cam_xpt_sim.h>
#include <cam/cam_debug.h>
#include <cam/scsi/scsi_all.h>
#include <cam/scsi/scsi_message.h>
#endif

#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/pmap.h>

/* Tell the driver memory allocator about this complexity */
#ifdef	FreeBSD_Bus_Dma_Abstraction
#define	SYM_OPT_BUS_DMA_ABSTRACTION
#endif

/* Short and quite clear integer types */
typedef int8_t    s8;
typedef int16_t   s16;
typedef	int32_t   s32;
typedef u_int8_t  u8;
typedef u_int16_t u16;
typedef	u_int32_t u32;

/*
 *  General driver includes.
 */
#include <dev/sym/sym_misc.h>
#include <dev/sym/sym_conf.h>
#include <dev/sym/sym_defs.h>

/*
 *  Insert a delay in micro-seconds and milli-seconds.
 */
void sym_udelay(int us);
void sym_mdelay(int ms);

/*
 *  Let the compiler know about driver data structure names.
 */
typedef struct sym_tcb *tcb_p;
typedef struct sym_lcb *lcb_p;
typedef struct sym_ccb *ccb_p;
typedef struct sym_hcb *hcb_p;
typedef struct sym_stcb *stcb_p;
typedef struct sym_slcb *slcb_p;
typedef struct sym_sccb *sccb_p;
typedef struct sym_shcb *shcb_p;

/*
 *  Define a reference to the O/S dependant IO request.
 */
typedef union ccb *cam_ccb_p;		/* Generic  */
typedef struct ccb_scsiio *cam_scsiio_p;/* SCSI I/O */

/*
 *  Some provision for a possible big endian mode supported by 
 *  Symbios chips (never seen, by the way).
 *  For now, this stuff does not deserve any comments. :)
 */
#define sym_offb(o)	(o)
#define sym_offw(o)	(o)

/*
 *  Some provision for support for BIG ENDIAN CPU.
 *  Btw, FreeBSD does not seem to be ready yet for big endian.
 */
#if	BYTE_ORDER == BIG_ENDIAN
#define cpu_to_scr(dw)	_htole32(dw)
#define scr_to_cpu(dw)	_le32toh(dw)
#else
#define cpu_to_scr(dw)	(dw)
#define scr_to_cpu(dw)	(dw)
#endif

/*
 *  Access to the chip IO registers and on-chip RAM.
 *  We use the `bus space' interface under FreeBSD-4 and 
 *  later kernel versions.
 */

#ifdef	FreeBSD_Bus_Space_Abstraction

#if defined(SYM_CONF_IOMAPPED)

#define INB_OFF(o)	bus_space_read_1(np->s.io.tag, np->s.io.bsh, o)
#define INW_OFF(o)	bus_space_read_2(np->s.io.tag, np->s.io.bsh, o)
#define INL_OFF(o)	bus_space_read_4(np->s.io.tag, np->s.io.bsh, o)

#define OUTB_OFF(o, v)	bus_space_write_1(np->s.io.tag, np->s.io.bsh, o, (v))
#define OUTW_OFF(o, v)	bus_space_write_2(np->s.io.tag, np->s.io.bsh, o, (v))
#define OUTL_OFF(o, v)	bus_space_write_4(np->s.io.tag, np->s.io.bsh, o, (v))

#else	/* Memory mapped IO */

#define INB_OFF(o)	bus_space_read_1(np->s.mmio.tag, np->s.mmio.bsh, o)
#define INW_OFF(o)	bus_space_read_2(np->s.mmio.tag, np->s.mmio.bsh, o)
#define INL_OFF(o)	bus_space_read_4(np->s.mmio.tag, np->s.mmio.bsh, o)

#define OUTB_OFF(o, v) bus_space_write_1(np->s.mmio.tag, np->s.mmio.bsh, o, (v))
#define OUTW_OFF(o, v) bus_space_write_2(np->s.mmio.tag, np->s.mmio.bsh, o, (v))
#define OUTL_OFF(o, v) bus_space_write_4(np->s.mmio.tag, np->s.mmio.bsh, o, (v))

#endif	/* SYM_CONF_IOMAPPED */

#define OUTRAM_OFF(o, a, l)	\
	bus_space_write_region_1(np->s.ram.tag, np->s.ram.bsh, o, (a), (l))

#else	/* not defined FreeBSD_Bus_Space_Abstraction */

#if	BYTE_ORDER == BIG_ENDIAN
#error	"BIG ENDIAN support requires bus space kernel interface"
#endif

/*
 *  Access to the chip IO registers and on-chip RAM.
 *  We use legacy MMIO and IO interface for FreeBSD 3.X versions.
 */

/*
 *  Define some understable verbs for IO and MMIO.
 */
#define io_read8(p)	 scr_to_cpu(inb((p)))
#define	io_read16(p)	 scr_to_cpu(inw((p)))
#define io_read32(p)	 scr_to_cpu(inl((p)))
#define	io_write8(p, v)	 outb((p), cpu_to_scr(v))
#define io_write16(p, v) outw((p), cpu_to_scr(v))
#define io_write32(p, v) outl((p), cpu_to_scr(v))

#ifdef	__alpha__

#define mmio_read8(a)	     readb(a)
#define mmio_read16(a)	     readw(a)
#define mmio_read32(a)	     readl(a)
#define mmio_write8(a, b)    writeb(a, b)
#define mmio_write16(a, b)   writew(a, b)
#define mmio_write32(a, b)   writel(a, b)
#define memcpy_to_pci(d, s, n)	memcpy_toio((u32)(d), (void *)(s), (n))

#else /*__i386__*/

#define mmio_read8(a)	     scr_to_cpu((*(volatile unsigned char *) (a)))
#define mmio_read16(a)	     scr_to_cpu((*(volatile unsigned short *) (a)))
#define mmio_read32(a)	     scr_to_cpu((*(volatile unsigned int *) (a)))
#define mmio_write8(a, b)   (*(volatile unsigned char *) (a)) = cpu_to_scr(b)
#define mmio_write16(a, b)  (*(volatile unsigned short *) (a)) = cpu_to_scr(b)
#define mmio_write32(a, b)  (*(volatile unsigned int *) (a)) = cpu_to_scr(b)
#define memcpy_to_pci(d, s, n)	bcopy((s), (void *)(d), (n))

#endif

/*
 *  Normal IO
 */
#if defined(SYM_CONF_IOMAPPED)

#define	INB_OFF(o)	io_read8(np->s.io.port + sym_offb(o))
#define	OUTB_OFF(o, v)	io_write8(np->s.io.port + sym_offb(o), (v))

#define	INW_OFF(o)	io_read16(np->s.io.port + sym_offw(o))
#define	OUTW_OFF(o, v)	io_write16(np->s.io.port + sym_offw(o), (v))

#define	INL_OFF(o)	io_read32(np->s.io.port + (o))
#define	OUTL_OFF(o, v)	io_write32(np->s.io.port + (o), (v))

#else	/* Memory mapped IO */

#define	INB_OFF(o)	mmio_read8(np->s.mmio.va + sym_offb(o))
#define	OUTB_OFF(o, v)	mmio_write8(np->s.mmio.va + sym_offb(o), (v))

#define	INW_OFF(o)	mmio_read16(np->s.mmio.va + sym_offw(o))
#define	OUTW_OFF(o, v)	mmio_write16(np->s.mmio.va + sym_offw(o), (v))

#define	INL_OFF(o)	mmio_read32(np->s.mmio.va + (o))
#define	OUTL_OFF(o, v)	mmio_write32(np->s.mmio.va + (o), (v))

#endif

#define OUTRAM_OFF(o, a, l) memcpy_to_pci(np->s.ram.va + (o), (a), (l))

#endif	/* FreeBSD_Bus_Space_Abstraction */

/*
 *  Misc. constants.
 */
#define SYM_SNS_BBUF_LEN	sizeof(struct scsi_sense_data)

/*
 *  Back-pointer from the CAM CCB to our data structures.
 */
#define sym_hcb_ptr	spriv_ptr0
/* #define sym_ccb_ptr	spriv_ptr1 */

/*
 *  System specific target data structure.
 *  Not needed under FreeBSD.
 */
/* struct sym_stcb */

/*
 *  System specific lun data structure.
 *  Not needed undee FreeBSD.
 */
/* struct sym_slcb */

/*
 *  System specific command data structure.
 */
#ifdef	FreeBSD_Bus_Dma_Abstraction
#define	SYM_HAVE_SCCB
struct sym_sccb {
	void		*arg;		/* Argument for some callback	*/
	bus_dmamap_t	dmamap;		/* DMA map for user data	*/
	u_char		dmamapped;	/* Direction we mapped it for	*/
#define SYM_DMA_NONE	0
#define SYM_DMA_READ	1
#define SYM_DMA_WRITE	2
};
#endif

/*
 *  Structures that host the mapping of IO and MMIO resources.
 */
#ifdef FreeBSD_Bus_Io_Abstraction
typedef device_t sym_device_t;
struct sym_bus_mem {
	int reg;
	struct resource	   *res;
	bus_space_handle_t bsh;
	bus_space_tag_t    tag;
};
struct sym_bus_io {
	int reg;
	struct resource	   *res;
	bus_space_handle_t bsh;
	bus_space_tag_t    tag;
};
#else
typedef pcici_t sym_device_t;
struct sym_bus_mem {
	vm_offset_t va;
};
struct sym_bus_io {
	pci_port_t port;
};
#endif

/*
 *  System specific host data structure.
 */
struct sym_shcb {
	/*
	 *  Chip and controller indentification.
	 */
	int	unit;
	char	inst_name[8];
	sym_device_t device;

	/*
	 *  CAM SIM information for this instance.
	 */
	struct cam_sim  *sim;
	struct cam_path *path;

	/*
	 *  Allocated hardware resources.
	 */
	struct sym_bus_mem mmio;
	struct sym_bus_mem ram;
	struct sym_bus_io  io;

#ifdef FreeBSD_Bus_Io_Abstraction
	struct resource	*irq_res;
	void *intr;
#endif

	/*
	 *  DMA stuff.
	 */
#ifdef	FreeBSD_Bus_Dma_Abstraction
	bus_dma_tag_t	data_dmat;	/* Tag for user data DMA	*/
#endif
	/*
	 *  CAM CCB pending queue.
	 */
	SYM_QUEHEAD	cam_ccbq;
};

/*
 *  Return the name of the controller.
 */
#define	sym_name(np) (np)->s.inst_name

/*
 *  Data structure pointer used as input for the NVRAM reading.
 *  Must resolve the IO macros and sym_name().
 *  The host data structure fits just fine under FreeBSD.
 */
typedef hcb_p sdev_p;

/*
 *  The driver definitions (sym_hipd.h) must know about a 
 *  couple of things related to the memory allocator.
 */
typedef u_long m_addr_t;	/* Enough bits to represent any address */
#define SYM_MEM_PAGE_ORDER 0	/* 1 PAGE  maximum */
#define SYM_MEM_CLUSTER_SHIFT	(PAGE_SHIFT+SYM_MEM_PAGE_ORDER)
#if 0
#define SYM_MEM_FREE_UNUSED	/* Free unused pages immediately */
#endif

#ifdef	FreeBSD_Bus_Dma_Abstraction
typedef bus_dma_tag_t m_pool_ident_t;

#define SYM_HAVE_M_SVTOB
struct sym_m_svtob {
	bus_dmamap_t dmamap;	/* Dma Map for this chunk */
};

#define SYM_HAVE_M_SPOOL
struct sym_m_spool {
	bus_dma_tag_t dmat;	/* Tag for our fixed allocations */
};
#endif

/*
 *  Include driver soft definitions.
 */
#include <dev/sym/sym_fw.h>
#include <dev/sym/sym_hipd.h>

/*
 *  Memory allocator.
 */
#define SYM_MEM_WARN	1	/* Warn on failed operations */

#define sym_get_mem_cluster()	malloc(SYM_MEM_CLUSTER_SIZE, M_DEVBUF, M_NOWAIT)
#define sym_free_mem_cluster(p)	free((p), M_DEVBUF)

#define	sym_calloc(size, name)	 sym_calloc_unlocked(size, name)
#define	sym_mfree(m, size, name) sym_mfree_unlocked(m, size, name)

#ifndef	FreeBSD_Bus_Dma_Abstraction
/*
 *  Simple case.
 *  All the memory assummed DMAable and O/S providing virtual 
 *  to bus physical address translation.
 */
#define __sym_calloc_dma(pool_id, size, name)	sym_calloc(size, name)
#define __sym_mfree_dma(pool_id, m, size, name)	sym_mfree( m, size, name)
#ifdef	__alpha__
#define	__vtobus(b, p)	alpha_XXX_dmamap((vm_offset_t)(p))
#else /*__i386__*/
#define __vtobus(b, p)	vtophys(p)
#endif

#else	/* FreeBSD_Bus_Dma_Abstraction */
/*
 *  Complex case.
 *  We have to provide the driver memory allocator with methods for 
 *  it to maintain virtual to bus physical address translations.
 */

#define sym_m_pool_match(mp_id1, mp_id2)	(mp_id1 == mp_id2)

void sym_cb_dmamap_load(void *arg, bus_dma_segment_t *segs, int nseg, int err);

static __inline m_addr_t sym_m_get_dma_mem_cluster(m_pool_p mp, m_vtob_p vbp)
{
	void *vaddr = 0;
	bus_addr_t baddr = 0;

	if (bus_dmamem_alloc(mp->s.dmat, &vaddr,
			      BUS_DMA_NOWAIT, &vbp->s.dmamap))
		goto out_err;
	bus_dmamap_load(mp->s.dmat, vbp->s.dmamap, vaddr,
			SYM_MEM_CLUSTER_SIZE, sym_cb_dmamap_load, &baddr, 0);
	if (baddr) {
		vbp->vaddr = (m_addr_t) vaddr;
		vbp->baddr = (m_addr_t) baddr;
		return (m_addr_t) vaddr;
	}
out_err:
	if (baddr)
		bus_dmamap_unload(mp->s.dmat, vbp->s.dmamap);
	if (vaddr)
		bus_dmamem_free(mp->s.dmat, vaddr, vbp->s.dmamap);
	return 0;
}

static __inline void sym_m_free_dma_mem_cluster(m_pool_p mp, m_vtob_p vbp)
{
	bus_dmamap_unload(mp->s.dmat, vbp->s.dmamap);
	bus_dmamem_free(mp->s.dmat, (void *) vbp->vaddr, vbp->s.dmamap);
	bus_dmamap_destroy(mp->s.dmat, vbp->s.dmamap);
}

static __inline int sym_m_create_dma_mem_tag(m_pool_p mp)
{
	return bus_dma_tag_create(mp->dev_dmat, 1, SYM_MEM_CLUSTER_SIZE,
		       BUS_SPACE_MAXADDR_32BIT,
		       BUS_SPACE_MAXADDR_32BIT,
		       NULL, NULL, SYM_MEM_CLUSTER_SIZE, 1,
		       SYM_MEM_CLUSTER_SIZE, 0, &mp->s.dmat);
}

static __inline void sym_m_delete_dma_mem_tag(m_pool_p mp)
{
	bus_dma_tag_destroy(mp->s.dmat);
}

#define __sym_calloc_dma(dev_dmat, size, name)		\
		__sym_calloc_dma_unlocked(dev_dmat, size, name)
#define __sym_mfree_dma(dev_dmat, m, size, name)	\
		__sym_mfree_dma_unlocked(dev_dmat, m, size, name)
#define __vtobus(dev_dmat, m)	\
		__vtobus_unlocked(dev_dmat, m)

#endif	/* FreeBSD_Bus_Dma_Abstraction */

/*
 *  Set the status field of a CAM CCB.
 */
static __inline void 
sym_set_cam_status(union ccb *ccb, cam_status status)
{
	ccb->ccb_h.status &= ~CAM_STATUS_MASK;
	ccb->ccb_h.status |= status;
}

/*
 *  Get the status field of a CAM CCB.
 */
static __inline int 
sym_get_cam_status(union ccb *ccb)
{
	return ccb->ccb_h.status & CAM_STATUS_MASK;
}

#ifdef	FreeBSD_Bus_Dma_Abstraction
/*
 *  Create a dma map for this IO.
 */
static __inline int  
sym_data_dmamap_create(hcb_p np, ccb_p cp)
{
	return bus_dmamap_create(np->s.data_dmat, 0, &cp->s.dmamap);
}
/*
 *  Destroy the dma map of this IO.
 */
static __inline void  
sym_data_dmamap_destroy(hcb_p np, ccb_p cp)
{
	bus_dmamap_destroy(np->s.data_dmat, cp->s.dmamap);
}
/*
 *  Unload the dma map for this IO.
 */
static __inline void 
sym_data_dmamap_unload(hcb_p np, ccb_p cp)
{
	if (cp->s.dmamapped) {
		bus_dmamap_unload(np->s.data_dmat, cp->s.dmamap);
		cp->s.dmamapped = 0;
	}
}
static __inline void 
sym_data_dmamap_presync(hcb_p np, ccb_p cp)
{
	if (cp->s.dmamapped) {
		bus_dmamap_sync(np->s.data_dmat, cp->s.dmamap,
			(bus_dmasync_op_t)(cp->s.dmamapped == SYM_DMA_READ ?
				BUS_DMASYNC_PREREAD : BUS_DMASYNC_PREWRITE));
	}
}
static __inline void 
sym_data_dmamap_postsync(hcb_p np, ccb_p cp)
{
	if (cp->s.dmamapped) {
		bus_dmamap_sync(np->s.data_dmat, cp->s.dmamap,
			(bus_dmasync_op_t)(cp->s.dmamapped == SYM_DMA_READ ?
				BUS_DMASYNC_POSTREAD : BUS_DMASYNC_POSTWRITE));	
	}
}
#else
#define sym_data_dmamap_create(np, cp)		(0)
#define sym_data_dmamap_destroy(np, cp)		do { ; } while (0)
#define sym_data_dmamap_unload(np, cp)		do { ; } while (0)
#define sym_data_dmamap_presync(np, cp)		do { ; } while (0)
#define sym_data_dmamap_postsync(np, cp)	do { ; } while (0)
#endif

/*
 *  Asynchronous callback on data transfer negotiations.
 */
void __sym_xpt_async_transfer_neg(hcb_p np, int target, u_int spi_valid);

#ifdef	FreeBSD_New_Tran_Settings
#define SYM_SPI_VALID_WDTR		\
	CTS_SPI_VALID_BUS_WIDTH |	\
	CTS_SPI_VALID_SYNC_RATE |	\
	CTS_SPI_VALID_SYNC_OFFSET
#define SYM_SPI_VALID_SDTR		\
	CTS_SPI_VALID_SYNC_RATE |	\
	CTS_SPI_VALID_SYNC_OFFSET
#define SYM_SPI_VALID_PPR		\
	CTS_SPI_VALID_PPR_OPTIONS |	\
	CTS_SPI_VALID_BUS_WIDTH |	\
	CTS_SPI_VALID_SYNC_RATE |	\
	CTS_SPI_VALID_SYNC_OFFSET
#else
#define SYM_SPI_VALID_WDTR		\
	CCB_TRANS_BUS_WIDTH_VALID |	\
	CCB_TRANS_SYNC_RATE_VALID |	\
	CCB_TRANS_SYNC_OFFSET_VALID
#define SYM_SPI_VALID_SDTR		\
	CCB_TRANS_SYNC_RATE_VALID |	\
	CCB_TRANS_SYNC_OFFSET_VALID
#define SYM_SPI_VALID_PPR		\
	CCB_TRANS_BUS_WIDTH_VALID |	\
	CCB_TRANS_SYNC_RATE_VALID |	\
	CCB_TRANS_SYNC_OFFSET_VALID
#endif

#define sym_xpt_async_nego_wide(np, target)	\
	__sym_xpt_async_transfer_neg(np, target, SYM_SPI_VALID_WDTR)
#define sym_xpt_async_nego_sync(np, target)	\
	__sym_xpt_async_transfer_neg(np, target, SYM_SPI_VALID_SDTR)
#define sym_xpt_async_nego_ppr(np, target)	\
	__sym_xpt_async_transfer_neg(np, target, SYM_SPI_VALID_PPR)

/*
 *  Build CAM result for a successful IO and for a failed IO.
 */
static __inline void sym_set_cam_result_ok(hcb_p np, ccb_p cp, int resid)
{
	struct ccb_scsiio *csio = &cp->cam_ccb->csio;

	csio->resid = resid;
	csio->scsi_status = cp->ssss_status;
	sym_set_cam_status((union ccb *) csio, CAM_REQ_CMP);
}
void sym_set_cam_result_error(hcb_p np, ccb_p cp, int resid);

/*
 *  Other O/S specific methods.
 */
#define sym_cam_target_id(ccb)	(ccb)->ccb_h.target_id
#define sym_cam_target_lun(ccb)	(ccb)->ccb_h.target_lun
void sym_freeze_cam_ccb(cam_ccb_p ccb);
void sym_xpt_done(hcb_p np, cam_ccb_p ccb);
void sym_xpt_done2(hcb_p np, cam_ccb_p ccb, int cam_status);
void sym_print_addr (ccb_p cp);
void sym_xpt_async_bus_reset(hcb_p np);
void sym_xpt_async_sent_bdr(hcb_p np, int target);
int  sym_setup_data_and_start (hcb_p np, struct ccb_scsiio *csio, ccb_p cp);
void sym_log_bus_error(hcb_p np);

#endif /* SYM_GLUE_H */
